#include "csetdata.c"
#include "cset.hh"

static const int EOL = -1;

static void SeqUTF8(std::string& result, wchar_t ch)
{
    //fprintf(stderr, "Sequencing %X\n", ch);
    const unsigned n = ch;
    if(n < 0x80)                 // <=7 bits
        result += (char)n;
    else
    {
        if(n < 0x800)            // <=11 bits
            result += (char)(0xC0 + (n>>6));
        else
        {
            if(n < 0x10000)      // <=16 bits
                result += (char)(0xE0 + (n>>12));
            else                 // <=21 bits
            {
                result += (char)(0xF0 + (n>>18));
                result += (char)(0x80 + ((n>>12)&63));
            }
            result += (char)(0x80 + ((n>>6)&63));
        }
        result += (char)(0x80 + (n&63));
    }
}
static void SeqLatin1(std::string& result, wchar_t ch)
{
    result += (char)ch;
}
static void SeqBeginEUC(std::string& result)
{
    result += "$B";
}
static void SeqEndEUC(std::string& result)
{
    result += "(B";
}
static void SeqEUC(std::string& result, unsigned jisvalue)
{
    char ch1 = (jisvalue >> 8) & 255;
    char ch2 = (jisvalue     ) & 255;
    result += ch1;
    result += ch2;
}
unsigned GetEUCvalue(wchar_t ch)
{
    unsigned unicode = ch;
    if(unicode >= min_uni_euc && unicode <= max_uni_euc)
        return uni_euc[unicode-min_uni_euc];
    return 0;
}
unsigned UngetEUCvalue(unsigned euc)
{
    unsigned value = euc;
    if(value >= min_euc && value <= max_euc)
        return euc_uni[value-min_euc];
    return 0;
}
unsigned UngetSJISvalue(unsigned sjis)
{
    unsigned value = sjis;
    if(value >= min_sjis && value <= max_sjis)
        return sjis_uni[value-min_sjis];
    return 0;
}

/*
 * Generate UTF-8
 */
const std::string toUTF8(const std::wstring& s)
{
    std::string result;
    result.reserve(s.size() * 3);
    for(unsigned a=0; a<s.size(); ++a)
        SeqUTF8(result, s[a]);
    return result;
}

/*
 * Generate ISO-8859-1
 * Fallback to ISO-2022-JP
 * Fallback to UTF-8
 *
 */
const std::string toLatin1(const std::wstring& s)
{
    std::string result;
    result.reserve(s.size() * 2);
    enum { latin1, jis, utf8 } mode = latin1;
    for(unsigned a=0; a<s.size(); ++a)
    {
        unsigned n = s[a];
        if(n < 256)
        {
            if(mode == jis) SeqEndEUC(result);
            SeqLatin1(result, n);
            mode = latin1;
        }
        else
        {
            unsigned JisValue = GetEUCvalue(n);
            if(JisValue > 0)
            {
                if(mode != jis) SeqBeginEUC(result);
                SeqEUC(result, JisValue);
                mode = jis;
            }
            else
            {
                if(mode == jis) SeqEndEUC(result);
                SeqUTF8(result, n);
                mode = utf8;
            }
        }
    }
    if(mode == jis) SeqEndEUC(result);
    return result;
}


/*
 * Read UTF8
 * Fallback to ISO-8859-1
 */
const std::wstring fromUTF8(const std::string& s)
{
    std::wstring result;
    result.reserve(s.size());
    for(unsigned a=0; a<s.size(); ++a)
    {
        int b0 = (unsigned char)s[a];
        int b1 = a+1 < s.size() ? (unsigned char)s[a+1] : EOL;
        int b2 = a+2 < s.size() ? (unsigned char)s[a+2] : EOL;
        int b3 = a+3 < s.size() ? (unsigned char)s[a+3] : EOL;
        if(b0 < 0x80)
        {
            // ascii
            result += b0;
        }
        else if(b1 != EOL
             && b0 >= 0xC1 && b0 < 0xE0
             && b1 >= 0x80 && b1 < 0xC0)
        {
            ++a; // 2-byte utf-8
            result += (b1 & 63) | ((b0 & 0x1F) << 6);
        }
        else if(b1 != EOL && b2 != EOL
             && b0 >= 0xE1 && b0 < 0xF0
             && b1 >= 0x80 && b1 < 0xC0
             && b2 >= 0x80 && b2 < 0xC0)
        {
            a += 2; // 3-byte utf-8
            result += (b2 & 63) | ((b1 & 63) << 6) | ((b0 & 0x0F) << 12);
        }
        else if(b1 != EOL && b2 != EOL && b3 != EOL
             && b0 >= 0xF1 && b0 < 0xF8
             && b1 >= 0x80 && b1 < 0xC0
             && b2 >= 0x80 && b2 < 0xC0
             && b3 >= 0x80 && b3 < 0xC0)
        {
            a += 3; // 4-byte utf-8
            result += (b3 & 63) | ((b2 & 63) << 6) | ((b1 & 63) << 12) | ((b0 & 0x07) << 16);
        }
        else
        {
            // fallback to iso-8859-1
            result += b0;
        }
    }
    return result;
}

/*
 * Read ISO-2022-JP
 * Fallback to SHIFTJIS
 * Fallback to UTF-8
 * Fallback to ISO-8859-1
 */
const std::wstring fromLatin1(const std::string& s)
{
    std::wstring result;
    result.reserve(s.size());
    
    enum { jis, guess } mode = guess;
    
    for(unsigned a=0; a<s.size(); ++a)
    {
        int b0 = (unsigned char)s[a];
        int b1 = a+1 < s.size() ? (unsigned char)s[a+1] : EOL;
        int b2 = a+2 < s.size() ? (unsigned char)s[a+2] : EOL;
        int b3 = a+3 < s.size() ? (unsigned char)s[a+3] : EOL;

        if(mode == jis)
        {
            if(b1 != EOF)
            {
                unsigned euc = (b0 << 8) | b1;
                unsigned uni = UngetEUCvalue(euc);
                if(uni != 0)
                {
                    ++a; // 2-byte iso-2022-jp
                    result += (wchar_t)uni;
                    continue;
                }
            }
            if(b0 == '' && b1 == '(' && b2 == 'B')
            {
                a += 2; // switch to ascii
                mode = guess;
                continue;
            }
        }
        mode = guess;
        if(b0 == '' && b1 == '$' && b2 == 'B')
        {
            a += 2; // switch to jis
            mode = jis;
            continue;
        }
        if(b0 >= 0x80 && b0 <= 0x9F
        && b1 >= 0x40 && b1 <= 0xFC)
        {
            unsigned sjis = (b0 << 8) | b1;
            unsigned uni = UngetSJISvalue(sjis);
            if(uni != 0)
            {
                ++a; // 2-byte sjis
                result += (wchar_t)uni;
                continue;
            }
        }
        if(b0 < 0x80)
        {
            // ascii
            result += b0;
        }
        else if(b1 != EOL
             && b0 >= 0xC1 && b0 < 0xE0
             && b1 >= 0x80 && b1 < 0xC0)
        {
            ++a; // 2-byte utf-8
            result += (b1 & 63) | ((b0 & 0x1F) << 6);
        }
        else if(b1 != EOL && b2 != EOL
             && b0 >= 0xE1 && b0 < 0xF0
             && b1 >= 0x80 && b1 < 0xC0
             && b2 >= 0x80 && b2 < 0xC0)
        {
            a += 2; // 3-byte utf-8
            result += (b2 & 63) | ((b1 & 63) << 6) | ((b0 & 0x0F) << 12);
        }
        else if(b1 != EOL && b2 != EOL && b3 != EOL
             && b0 >= 0xF1 && b0 < 0xF8
             && b1 >= 0x80 && b1 < 0xC0
             && b2 >= 0x80 && b2 < 0xC0
             && b3 >= 0x80 && b3 < 0xC0)
        {
            a += 3; // 4-byte utf-8
            result += (b3 & 63) | ((b2 & 63) << 6) | ((b1 & 63) << 12) | ((b0 & 0x07) << 16);
        }
        else
        {
            // 2-byte sequence that contains controlcodes might be sjis
            if(b0 >= 0xE0 && b0 <= 0xEA
            && b1 >= 0x80 && b1 <= 0x9F)
            {
                unsigned sjis = (b0 << 8) | b1;
                unsigned uni = UngetSJISvalue(sjis);
                if(uni != 0)
                {
                    ++a; // 2-byte sjis
                    result += (wchar_t)uni;
                    continue;
                }
            }
            
            // fallback to iso-8859-1
            result += b0;
        }
    }
    return result;
}
